/* $Id: ldpc_fec.cpp,v 1.29 2004/01/23 14:07:53 chneuman Exp $ */

/*  LDPC/LDGM FEC Library for Java Native Interface (JNI).
 *  Call ldpc_fec.cpp
 *  (c) Copyright 2004 INRIA - All rights reserved
 *
 *  Main author : Sylvain Dattrino (s-datt02@univ-lyon1.fr)
 *                Cong-Dug Pham    (congduc.pham@ens-lyon.fr)
 *
 *  Main authors of ldpc_fec.cpp :     Julien Laboure  (julien.laboure@inrialpes.fr)
 *                                     Zainab Khallouf (zainab.khallouf@inrialpes.fr)
 *                                     Vincent Roca    (vincent.roca@inrialpes.fr)
 *                                     Christoph Neumann (christoph.neumann@inrialpes.fr)
 *
 *  This copyright notice must be retained and prominently displayed,
 *  along with a note saying that the original programs are available from
 *  Cong-Dug Pham's web page, and note is made of any changes made to these
 *  programs.  
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/time.h>
#include <netinet/in.h>

#include "LDPCFecSession_0005fJNI.h"
#include "../src/ldpc_fec.h"

/* Prototypes */

LDPCFecSession* Getjptr_obj(JNIEnv*, jobject);

/******************************************************************************
 * LDPCFecSession_JNI Contructor
 * => See java LDPCFecSession_JNI class file for more informations.
 */

JNIEXPORT jint JNICALL Java_LDPCFecSession_1JNI_LDPCFecSession
(JNIEnv *jenv, jobject jobject_jni)
{
  return ((jint) new LDPCFecSession());
}

/******************************************************************************
 * LDPCFecSession_JNI Destructor
 * => See java LDPCFecSession_JNI class file for more informations.
 */

/******************************************************************************
 * InitSession : Initializes the LDPC session.
 * => See java LDPCFecSession_JNI class file for more informations.
 */

JNIEXPORT jint JNICALL Java_LDPCFecSession_1JNI_InitSession
(JNIEnv *jenv_jni, jobject jobject_jni, jint jnbDataPkt, jint jnbFecPkt, jint jpktSize, jint jflags, jint jleftDegree, jint jseed, jint jcodecType)
{
  LDPCFecSession* jptr_obj=Getjptr_obj(jenv_jni,jobject_jni);

  return (int)(jptr_obj->InitSession((int)jnbDataPkt,(int)jnbFecPkt,(int)jpktSize,(int)jflags,(int)jleftDegree,(int)jseed,(SessionType)jcodecType));
}

/******************************************************************************
 * EndSession: Ends the LDPC session, cleans up everything.
 *
 * => See java LDPCFecSession_JNI class file for more informations.
 */

JNIEXPORT void JNICALL Java_LDPCFecSession_1JNI_EndSession
(JNIEnv *jenv_jni, jobject jobject_jni)
{

  LDPCFecSession* jptr_obj=Getjptr_obj(jenv_jni,jobject_jni);

  jptr_obj->EndSession();

  return;
}

/******************************************************************************
 * IsInitialized: Check if the LDPC session has been initialized.
 * 
 * => See java LDPCFecSession_JNI class file for more informations.
 */

JNIEXPORT jboolean JNICALL Java_LDPCFecSession_1JNI_IsInitialized
(JNIEnv *jenv_jni, jobject jobject_jni)
{
  LDPCFecSession* jptr_obj=Getjptr_obj(jenv_jni,jobject_jni);

  return (jboolean)(jptr_obj->IsInitialized());
}
  

/******************************************************************************
 * SetVerbosity: Sets the verbosity level.
 *
 * => See java LDPCFecSession_JNI class file for more informations.
 */

JNIEXPORT void JNICALL Java_LDPCFecSession_1JNI_SetVerbosity
(JNIEnv *jenv_jni, jobject jobject_jni, jint jverb)
{
  LDPCFecSession* jptr_obj=Getjptr_obj(jenv_jni,jobject_jni);

  jptr_obj->SetVerbosity((int)jverb);

  return;
}

/******************************************************************************
 * BuildFecPacket: Builds a new FEC packet.
 *
 * => See java LDPCFecSession_JNI class file for more informations.
 */

JNIEXPORT void JNICALL Java_LDPCFecSession_1JNI_BuildFecPacket
(JNIEnv *jenv_jni, jobject jobject_jni, jobjectArray jpkt_canvas, jint jfec_index, jbyteArray jfec_pkt)
{

  jbyteArray jPacket_Array=NULL;
  LDPCFecSession* jptr_obj;

  char** packetsArray=NULL;
  char*  packet=NULL;

  int i;

  int nbPkt=(int)jenv_jni->GetArrayLength(jpkt_canvas);     // cast from jsize to int

  int pktSize=(int)jenv_jni->GetArrayLength(jfec_pkt);     // cast from jsize to int

  jptr_obj=Getjptr_obj(jenv_jni,jobject_jni);

  packetsArray = (char**)calloc(nbPkt, sizeof(char*));

  if( packetsArray == NULL )
    {
      printf("Error: insufficient memory (calloc failed for packetsArray)\n");
      
      if(jptr_obj->IsInitialized())
	jptr_obj->EndSession();

      return;

    }

  for( i=0; i<nbPkt; i++ )
    {
      jPacket_Array=(jbyteArray)jenv_jni->GetObjectArrayElement(jpkt_canvas, i);

      if(jPacket_Array)
	packetsArray[i]=(char*) jenv_jni->GetByteArrayElements(jPacket_Array, 0);     // cast for jbtye* to char*
      else
	packetsArray[i]=NULL;     // flag for releasing

    }

  packet=(char*) jenv_jni->GetByteArrayElements(jfec_pkt, 0);     // cast for jbtye* to char*

  jptr_obj->BuildFecPacket((void**) packetsArray,(int) jfec_index,(void*) packet);

  // Commit changes to the Java Virtual Machine and free some memory

  jPacket_Array = jenv_jni->NewByteArray(pktSize);

  for (i=0; i<nbPkt; i++)
    if(packetsArray[i])     // utilization of flag
      {

        jenv_jni->ReleaseByteArrayElements(jPacket_Array,(jbyte*)packetsArray[i],0);

	jenv_jni->SetObjectArrayElement(jpkt_canvas,i,jPacket_Array);

      }

  jenv_jni->ReleaseByteArrayElements(jfec_pkt,(jbyte *)packet,0);

  return;
}

/******************************************************************************
 * DecodeFecStep: Perform a new decoding step with a new (given) packet.
 *
 * => See java LDPCFecSession_JNI class file for more informations.
 */

JNIEXPORT jint JNICALL Java_LDPCFecSession_1JNI_DecodeFecStep
(JNIEnv *jenv_jni, jobject jobject_jni, jobjectArray jpkt_canvas, jbyteArray jnew_packet, jint jpkt_seqno, jboolean jstore_packet, jint jcontext_4callback)
{
  jbyteArray jPacket_Array=NULL;
  LDPCFecSession* jptr_obj=NULL;

  char** packetsArray=NULL;
  char*  packet=NULL;

  int i;

  jint ret;

  int nbPkt=(int)jenv_jni->GetArrayLength(jpkt_canvas);     // cast from jsize to int
  
  // int pktSize=(int)jenv_jni->GetArrayLength(jnew_packet);     // cast from jsize to int

  jptr_obj=Getjptr_obj(jenv_jni,jobject_jni);

  packetsArray = (char**)calloc(nbPkt, sizeof(char*));

  if( packetsArray == NULL )
    {
      printf("Error: insufficient memory (calloc failed for packetsArray)\n");
      
      if(jptr_obj->IsInitialized())
	jptr_obj->EndSession();

      return (jint)LDPC_ERROR;

    }

  for( i=0; i<nbPkt; i++ )
    {
      jPacket_Array=(jbyteArray)jenv_jni->GetObjectArrayElement(jpkt_canvas, i);

      if(jPacket_Array)
	packetsArray[i]=(char*) jenv_jni->GetByteArrayElements(jPacket_Array, 0);     // cast for jbtye* to char*
      else
	packetsArray[i]=NULL;     // flag for releasing

    }

  packet=(char*) jenv_jni->GetByteArrayElements(jnew_packet, 0);     // cast for jbtye* to char*
  
  ret=(jint) (jptr_obj->DecodeFecStep((void**) packetsArray,(void*) packet,(int) jpkt_seqno, (bool) jstore_packet,(void*) jcontext_4callback));

  // Commit changes to the Java Virtual Machine

  // Looking for packets decoded, if there are one; and update

  for(i=0;i<nbPkt;i++)
    if(packetsArray[i])     // utilization of flag
      {
	jenv_jni->ReleaseByteArrayElements(jPacket_Array,(jbyte*)packetsArray[i],0);
	
	jenv_jni->SetObjectArrayElement(jpkt_canvas,i,jPacket_Array);
	
      }

  jenv_jni->ReleaseByteArrayElements(jnew_packet,(jbyte*)packet,0);
  
  return ret;

}

/******************************************************************************
 * IsDecodingComplete: Checks if all DATA packets have been received/rebuilt.
 *
 * => See java LDPCFecSession_JNI class file for more informations.
 */

JNIEXPORT jboolean JNICALL Java_LDPCFecSession_1JNI_IsDecodingComplete
(JNIEnv *jenv_jni, jobject jobject_jni, jobjectArray jpkt_canvas)
{
  jbyteArray jPacket_Array=NULL;
  LDPCFecSession* jptr_obj=NULL;

  char** packetsArray=NULL;

  int i;
  int nbPkt=(int)jenv_jni->GetArrayLength(jpkt_canvas);     // cast from jsize to int

  jboolean ret;

  jptr_obj=Getjptr_obj(jenv_jni,jobject_jni);

  packetsArray = (char**)calloc(nbPkt, sizeof(char*));

  if( packetsArray == NULL )
    {
      printf("Error: insufficient memory (calloc failed for packetsArray)\n");
      
      if(jptr_obj->IsInitialized())
	jptr_obj->EndSession();

      return false;

    }

  for( i=0; i<nbPkt; i++ )
    {
      jPacket_Array=(jbyteArray)jenv_jni->GetObjectArrayElement(jpkt_canvas, i);
      
      if(jPacket_Array)
	packetsArray[i]=(char*) jenv_jni->GetByteArrayElements(jPacket_Array, 0);     // cast for jbtye* to char*
    }

  ret = (jboolean) jptr_obj->IsDecodingComplete((void**) packetsArray);

  for(i=0;i<nbPkt;i++)
    if(packetsArray[i])     // utilization of flag
      {
	jenv_jni->ReleaseByteArrayElements(jPacket_Array,(jbyte*)packetsArray[i],0);
	
	jenv_jni->SetObjectArrayElement(jpkt_canvas,i,jPacket_Array);
	
      }

  return ret;

}

/******************************************************************************
 * Getjptr_obj: return pointer to the current LDPCFecSession
 */

LDPCFecSession* Getjptr_obj(JNIEnv* env, jobject jobject_jni)
{
  jclass cls;
  jfieldID fid;
  LDPCFecSession* jptr_obj;

  cls=env->GetObjectClass(jobject_jni);
  fid=env->GetFieldID(cls, "ptr_obj", "I");
  jptr_obj=(LDPCFecSession*)env->GetIntField(jobject_jni, fid);

  if(env->ExceptionOccurred())
    env->ExceptionClear();

  return jptr_obj;
}
