/* $Id: ldpc_fec.java,v 1.28 (translate from ldpc_fec.cpp  2003/12/17 11:55:39 roca Exp) $ */
/*
 *  LDPC/LDGM FEC Java Wrapper Library.
 *  Based on ldpc_fec.cpp
 *  Copyright (c) 2004 INRIA - All rights reserved
 *
 *  Main authors : Sylvain Dattrino (s-datt02@univ-lyon1.fr)
 *                 Cong-Dug Pham    (congduc.pham@ens-lyon.fr)
 *
 *  Main authors of ldpc_fec.cpp : Julien Laboure  (julien.laboure@inrialpes.fr)
 *                Zainab Khallouf (zainab.khallouf@inrialpes.fr)
 *                Vincent Roca    (vincent.roca@inrialpes.fr)
 *
 *  This copyright notice must be retained and prominently displayed,
 *  along with a note saying that the original programs are available from
 *  Cong-Dug Pham's web page, and note is made of any changes made to these
 *  programs.  
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * This is the FEC session class, where all the context information
 * is kept for encoding/decoding this block.
 * There must be one such FEC session instance per FEC block.
 * <p>
 * WARNING: the following class contains a lot of checking code that
 * is only available in DEBUG mode (set -DDEBUG on the compiling line
 * of native implementation code).<br>
 * Whenever used with a new application, first validate your code in
 * DEBUG mode, and switch to production code only in a second step...
 */

class LDPCFecSession_JNI
{
    
    // Error status returned by functions.

    /** Error statut retruned by functions*/
    public final static int LDPC_OK=0;
    /** Ok statut retruned by functions*/
    public final static int LDPC_ERROR=1;

    // Is the session a coding or decoding session, or both.

    /** Flag for a coding session*/
    public final static int FLAG_CODER=0x00000001;
    /** Flag for a decoding session*/
    public final static int FLAG_DECODER=0x00000002;
    /** Flag for a coding and decoding session*/
    public final static int FLAG_BOTH=(FLAG_CODER|FLAG_DECODER);

    // Type of codec to use for the session.

    /** Codec of type Low-Density Parity Check */
    public final static int TypeLDPC=0;
    /** Codec of type Low Density-Generator Matrix */
    public final static int TypeLDGM=1;
    /** Codec of type Stairs */
    public final static int TypeSTAIRS=2;
    /** Codec of type Triangle */
    public final static int TypeTRIANGLE=3;

    // Library of native code to load.

    static{System.loadLibrary("ldpc_fec");}

    /**
     * Pointer to an instance of native type LDPCFecSession
     * Public for JNI.
     */

    public int ptr_obj;

/******************************************************************************
 * LDPCFecSession_JNI Contructor.
 * <p>
 * => Call native LDPCFecSession Constructor and affect an int pointeur to the
 * member ptr_obj;
 */
    public LDPCFecSession_JNI()
    {
	ptr_obj=LDPCFecSession();
    }

    private native int LDPCFecSession();
    
/******************************************************************************
 * LDPCFecSession native Destructor.
 * <p>
 * => Call native LDPCFecSession Destructor and affect a zero pointeur to the
 * member ptr_obj;
 */
/*
    public ~LDPCFecSession_JNI()
    {
	~LDPCFecSession();
	ptr_obj=0;
    }

    private native void ~LDPCFecSession();

*/

/******************************************************************************
 * InitSession: Initializes the LDPC session.
 *
 * @param nbDataPkt	(IN) number of DATA packets (i.e. k).
 * @param nbFecPkt	(IN) number of FEC packets (i.e. n-k).
 * @param pktSize	(IN) packet size in bytes. MUST BE multiple of 4.
 * @param flags		(IN) session flags (FLAG_CODER, FLAG_DECODER, ...).
 * @param leftDegree	(IN) number of check edges per packet (FEC constraints).
 * @param seed		(IN) seed used to build the parity check matrix (H).
 * @param codecType	(IN) Type of codec algorithm and matrix to use.
 *			Can be on of TypeLDGM, TypeSTAIRS, TypeLDPC
 * @return		Completion status (LDPC_OK or LDPC_ERROR).
 */

    public native int InitSession       (int	nbDataPkt,
					 int	nbFecPkt,
					 int        pktSize,
					 int	flags,
					 int	leftDegree,
					 int	seed,
					 int	codecType);

/******************************************************************************
 * SetCallbackFunctions: Set the various callback functions for this session.
 * <p>
 * - The TxMalloc callback function is called each time the encoding function
 * needs to allocate memory for a new FEC packet.<br>
 * - The RxMalloc callback function is called each time the decoding function
 * needs to allocate memory for a new decoded packet.<br>
 * - The DecodedPkt callback function will be called each time a packet
 * is decoded by the DecodeFecStep() function. What this function does is
 * application-dependant.
 *
 * @param DecodedPkt_callback	(IN) Pointer to an application's callback.
 * 				This function will be given the size of
 * 				the newly created data or FEC packet,
 *				the sequence number of this packet, and
 * 				an opaque context parameter. This context
 * 				parameter is the same as that given to
 * 				DecodeFecStep(). This function will return
 *				a pointer to the data buffer allocated.
 * @return			Completion status (LDPC_OK or LDPC_ERROR).
 */

/*

public int SetCallbackFunctions (
	void* (*DecodedPkt_callback) (int size, int pkt_seqno, void* context))
{
	if (DecodedPkt_callback) {
		ASSERT(m_sessionFlags & FLAG_DECODER);
		m_decodedPkt_callback = DecodedPkt_callback;
	}
	return LDPC_OK;
}

*/


/******************************************************************************
 * EndSession : Ends the LDPC session, and cleans up everything.
 */

    public native void EndSession();

/******************************************************************************
 * IsInitialized: Check if the LDPC session has been initialized.
 *
 * @return	  TRUE if the session is ready and initialized, FALSE if not.
 */
    public native boolean  IsInitialized();


/******************************************************************************
 * Set the verbosity level.
 *
 * @param verb		(IN) new verbosity level (0: no trace, 1: few traces, 2: all traces)
 */

    public native void SetVerbosity(int verb);

/******************************************************************************
 * BuildFecPacket: Build a new FEC packet.
 *
 * @param pkt_canvas	(IN)    Array of source DATA packets.
 *				This is a table of n references to objects
 *				containing the source packets.
 * @param fec_index	(IN)	Index of FEC packet to build in {0.. n-k-1}
 *				(!) range.
 * @param fec_pkt	(IN-OUT)Reference to the FEC packet buffer that will
 *				be built. This buffer MUST BE allocated
 *				before,	but NOT cleared since
 *				this function will do it.
 */
	public native void BuildFecPacket(byte[][]	pkt_canvas,
					  int		fec_index,
					  byte[]	fec_pkt );

/******************************************************************************
 * DecodeFecStep: Perform a new decoding step with a new (given) packet.
 * <p>
 * This function relies on the following simple algorithm:<br>
 * <br>
 * Given a set of linear equations, if one of them has only one
 * remaining unknown variable, then the value of this variable is
 * that of the constant term.
 * Replace this variable by its value in all remaining linear
 * equations, and reiterate. The value of several variables can
 * therefore be find by this recursive algorithm.
 * <p>
 * In practice, an incoming packet contains the value of the associated
 * variable, so replace its value in all linear equations in which
 * it is implicated. Then apply the above algorithm and see if decoding
 * can progress by one or more steps.
 * <p>
 * For instance, if {s1, s2} are source symbols, and {f1} a FEC symbol:<br>
 *    { s1 + s2 + f1 = 0      (eq. 1)<br>
 *    {      s2 + f1 = 0      (eq. 2)<br>
 * <p>
 * Now if the node receives symbol s2, he replaces its value in the two
 * equations, he then finds f1, he replaces its value in the first equation
 * and finds s1.
 *
 * @param pkt_canvas	(IN-OUT) Global array of received/rebuilt packets.
 *				This is a table of n references to objects.
 * 				This array must be cleared upon
 * 				the first call to this function. It will be
 * 				automatically updated, with pointers to
 * 				packets received or decoded, by this function.
 * @param new_packet	(IN)	References to the buffer containing the new packet.
 * @param pkt_seqno	(IN)	Packet's sequence number in {0.. n-1} range.
 * @param store_packet	(IN)	true if the function needs to allocate memory,
 *				copy the packet content in it, and call
 *				any required callback.
 *				This is typically done when this function is
 *				called recursively, for newly decoded packets.
 * @param context_4_callback (IN) Pointer to context that will be passed
 * 				to the callback function (if any). This
 * 				context is not interpreted by this function.
 * @return			Completion status (LDPC_OK or LDPC_ERROR).
 */

    public native int DecodeFecStep(byte[][] pkt_canvas,
				    byte[]	new_packet,
				    int	pkt_seqno,
				    boolean	store_packet,
				    int context_4_callback);

/******************************************************************************
 * IsDecodingComplete: Checks if all DATA packets have been received/rebuilt.
 *
 * @param pkt_canvas	(IN)	Array of received/rebuilt packets.
 *
 * @return			TRUE if all DATA packets have been received
 * 				or decoded.
 */

    public native boolean IsDecodingComplete(byte[][] pkt_canvas);

}
