/* $Id: simple_coder.java coded from simple_coder.cpp,v 1.14 2003/09/23 16:30:15 laboure Exp $ */

/*  LDPC simple Java FEC encoder sample.
 *  Based on simple_coder.cpp
 *  (c) Copyright 2004 INRIA - All rights reserved
 *
 *  Main author : Sylvain Dattrino (s-datt02@univ-lyon1.fr)
 *                Cong-Dug Pham    (congduc.pham@ens-lyon.fr)
 *
 *  Main authors of simple_coder.cpp: Julien Laboure  (julien.laboure@inrialpes.fr)
 *                Zainab Khallouf (zainab.khallouf@inrialpes.fr)
 *                Vincent Roca    (vincent.roca@inrialpes.fr)
 *
 *  This copyright notice must be retained and prominently displayed,
 *  along with a note saying that the original programs are available from
 *  Cong-Dug Pham's web page, and note is made of any changes made to these
 *  programs.  
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

import java.net.*;
import java.io.*;
import java.util.*;

public class simple_coder
{

    // To synchronize with JNI

    final static int PKTSZ=1024;
    final static int NBDATA=50;
    final static int NBFEC=25;
    final static int NBPKT=NBDATA+NBFEC;
    final static int LEFT_DEGREE=3;

    final static int SESSION_TYPE=LDPCFecSession_JNI.TypeTRIANGLE;

    final static int SEED=2003;

    final static int VERBOSITY=1;

    static InetAddress DEST_IP;

    final static int DEST_PORT=10978;

    public static void main(String[] args)
    {
	simple_coder_java();

	return;

    }

    public static int simple_coder_java()
    {
	// Our session object...
	LDPCFecSession_JNI MyFecSession=new LDPCFecSession_JNI();

	
	// Original packets (DATA and FEC) are stored in a packets array
	// where each packet is an array of bytes
	byte[][] packetsArray;

    try
	{
	    DEST_IP=InetAddress.getByName("127.0.0.1");
	}
    catch(UnknownHostException e)
	{
	    System.out.println("Host Unknown");
	    return -1;
	}

    DatagramSocket mySock;
    ByteArrayOutputStream baos;
    DataOutputStream dos;
    DatagramPacket dp;

    int[] randOrder;
    int	ret		= -1;
    int	i,j;

    // Sets the verbosity level
    MyFecSession.SetVerbosity(VERBOSITY);

    if(MyFecSession.InitSession( NBDATA, NBFEC, PKTSZ, LDPCFecSession_JNI.FLAG_CODER, LEFT_DEGREE, SEED, SESSION_TYPE ) == LDPCFecSession_JNI.LDPC_ERROR)
	{
	    System.out.print("Error: Unable to initialize LDPC Session\n");
	    
	    DEST_IP=null;

	    return -1;
	}
	
    // Let's alloc our fictive DATA Packets...
    System.out.print("\nFilling DATA Packets...\n");
    try
	{
	    packetsArray = new byte[NBPKT][];
	}
    catch(OutOfMemoryError e)
	{
	    System.out.print("Error: insufficient memory (new failed for packetsArray)\n");

	    MyFecSession.EndSession();
	    DEST_IP=null;
	    MyFecSession=null;

	    return -1;
	}

    for( i=0; i<NBDATA; i++ )
	{	// First packet filled with 0x1111..., second with 0x2222..., etc.
	    try
		{
		    packetsArray[i] = new byte[PKTSZ];
		}
	    catch (OutOfMemoryError e)
		{
		    System.out.print("Error: insufficient memory (calloc failed for packetsArray["+i+"])\n");

		    MyFecSession.EndSession();
		    DEST_IP=null;
		    MyFecSession=null;

		    for(j=0;j<i;j++)
			packetsArray[i] = null;

		    packetsArray=null;

		    return -1;
		}

	    for (j=0;j<PKTSZ;j++)
		packetsArray[i][j]=(byte)(i+1);

	    if (VERBOSITY > 1)	// dump packet to screen (if wanted)
		{
		    System.out.print( "DATA["+i+"]= ");
		    DumpBuffer(packetsArray[i]);
		}
	}

    // Now creating some FEC Packets...
	System.out.print("\nCreating FEC Packets...\n");

	for( i=0; i < NBFEC; i++ )
	{

	    try
		{
		    packetsArray[i+NBDATA] = new byte[PKTSZ];
		}
	    catch(OutOfMemoryError e)
		{
		    System.out.print("Error: insufficient memory (calloc failed for packetsArray["+i+"])\n");

		    MyFecSession.EndSession();

		    DEST_IP=null;
		    MyFecSession=null;

		    for(j=0;j<i+NBDATA;j++)
			packetsArray[i] = null;

		    packetsArray=null;

		    return -1;
		}
		
	    MyFecSession.BuildFecPacket(packetsArray, i, packetsArray[i+NBDATA] );
		
	    if (VERBOSITY > 1)	// dump packet to screen (if wanted)
		{
		    System.out.print( "DATA["+(i+NBDATA)+"]= ");
		    DumpBuffer( packetsArray[i+NBDATA]);  // dump packet to screen
		}
	}


	// Randomize packets order...
	System.out.print("\nRandomizing transmit order...\n");

	try
    {
	randOrder = new int[NBPKT];
    }
    catch (OutOfMemoryError e)
    {
	System.out.print( "Error: insufficient memory (new failed for randOrder)\n" );

	MyFecSession.EndSession();

	DEST_IP=null;
	MyFecSession=null;

	for(j=0;j<NBPKT;j++)
	    packetsArray[i] = null;

	packetsArray=null;

	return -1;
    }
	
    randomizeArray(randOrder);

    // ... and finally, throw our packets to space using UDP socket :-)

    int sizeof_int=4;

    byte[] pktWithHeader = new byte[PKTSZ+sizeof_int];

    try
    {
	mySock = new DatagramSocket();
    }
    catch (Exception e)
    {
	System.out.print( "simple_coder_java: Error initializing socket!\n" );
	
	MyFecSession.EndSession();

	DEST_IP=null;
	MyFecSession=null;

	for(j=0;j<NBPKT;j++)
	    packetsArray[i] = null;

	packetsArray=null;

	randOrder=null;

	return -1;
    }
	
    System.out.print( "Sending packets (DATA&FEC) to "+DEST_IP+"/"+DEST_PORT+"\n");

    for( i=0; i < NBPKT; i++ )
    {
	// Adding pkt header (wich only countains a 32bits sequence number)

	try
	    {
		baos = new ByteArrayOutputStream ();

		dos = new DataOutputStream (baos);
	    }
	catch(OutOfMemoryError e)
	    {
		System.out.print( "Error: insufficient memory (new failed for ByteArrayOutputStream or DataOutputStream)\n" );

		MyFecSession.EndSession();

		DEST_IP=null;
		MyFecSession=null;

		for(j=0;j<NBPKT;j++)
		    packetsArray[i] = null;

		packetsArray=null;
		randOrder=null;
		mySock=null;
		baos=null;

		return -1;
	    }
	
	try
	    {
		dos.writeInt(randOrder[i]);
	    }
	catch(IOException e)
	    {
		System.out.println("IOException during conversion of a int !");

		MyFecSession.EndSession();

		DEST_IP=null;
		MyFecSession=null;

		for(j=0;j<NBPKT;j++)
		    packetsArray[i] = null;

		packetsArray=null;
		randOrder=null;
		mySock=null;
		baos=null;
		dos=null;

		return -1;
	    }

	byte[] identifiant=baos.toByteArray();

	for (j=0;j<identifiant.length;j++)
	    pktWithHeader[j]=identifiant[j];

	for (j=0;j<PKTSZ;j++)
	    pktWithHeader[j+4]=packetsArray[randOrder[i]][j];
		    
	System.out.print(""+(i+1)+"=> Sending packet "+randOrder[i]+" ("+(randOrder[i]<NBDATA ? "DATA" : "FEC")+")\n");

	try
	    { 
		dp= new DatagramPacket(pktWithHeader,pktWithHeader.length,DEST_IP,DEST_PORT);
	    }
	catch(OutOfMemoryError e)
	    {
		System.out.print( "Error: insufficient memory (new failed for DatagramPacket)\n" );

		MyFecSession.EndSession();

		DEST_IP=null;
		MyFecSession=null;
		
		for(j=0;j<NBPKT;j++)
		    packetsArray[i] = null;

		packetsArray=null;
		randOrder=null;
		mySock=null;
		baos=null;
		dos=null;

		return -1;
	    }

	try
	    {
		mySock.send(dp);
	    }
	catch(IOException e)
	    {
		System.out.print( "simple_coder_java: Error! send() failed!\n" );

		MyFecSession.EndSession();

		DEST_IP=null;
		MyFecSession=null;
		
		for(j=0;j<NBPKT;j++)
		    packetsArray[i] = null;

		packetsArray=null;
		randOrder=null;
		mySock=null;
		baos=null;
		dos=null;
		dp=null;

		ret = -1;
		break;
	    }
	try
	    {
		Thread.sleep(10); // sleep avoid UDP flood (in milliseconds)
	    }
	catch (InterruptedException e)
	    {
		// not so important
	    }
    }
    if( i==NBPKT ) {
	System.out.print( "\nComplete! "+i+" packets sent successfully.\n");
	ret = 1;
    }

    mySock.close();

    MyFecSession.EndSession();

    DEST_IP=null;
    MyFecSession=null;
		
    for(j=0;j<NBPKT;j++)
	packetsArray[j] = null;
    
    packetsArray=null;
    randOrder=null;
    mySock=null;
    baos=null;
    dos=null;
    dp=null;

    return ret;

}

// Randomize an array of integers
public static void randomizeArray(int[] array)
{
    int backup,randInd,i;

    int arrayLen=array.length;
    Random randomSequence = new Random();     // Its seed is initialized to a value based on the current time

    for( i=0; i<arrayLen; i++ )
	array[i]=i;

    for( i=0; i<arrayLen; i++ )
	{
	    backup = array[i];
	    randInd = randomSequence.nextInt(arrayLen);
	    array[i] = array[randInd];
	    array[randInd] = backup;
	}
}

public static void DumpBuffer(byte[] buff)
{
    int i;

    System.out.print("0x");

    for (i=0;i<buff.length;i++)
	System.out.print(""+buff[i]+":");

    System.out.print("\n");
}

}
