/* $Id: simple_decoder.java coded from simple_decoder.cpp,v 1.11 2003/09/23 16:30:15 laboure Exp $ */

/*  LDPC simple Java FEC decoder sample.
 *  Based on simple_decoder.cpp
 *  (c) Copyright 2004 INRIA - All rights reserved
 *
 *  Main author : Sylvain Dattrino (s-datt02@univ-lyon1.fr)
 *                Cong-Dug Pham    (congduc.pham@ens-lyon.fr)
 *
 *  Main authors of simple_decoder.cpp: Julien Laboure  (julien.laboure@inrialpes.fr)
 *                Zainab Khallouf (zainab.khallouf@inrialpes.fr)
 *                Vincent Roca    (vincent.roca@inrialpes.fr)
 *
 *  This copyright notice must be retained and prominently displayed,
 *  along with a note saying that the original programs are available from
 *  Cong-Dug Pham's web page, and note is made of any changes made to these
 *  programs.  
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

import java.net.*;
import java.io.*;
import java.util.*;

public class simple_decoder
{

    // To synchronize with JNI

    final static int PKTSZ=1024;
    final static int NBDATA=50;
    final static int NBFEC=25;
    final static int NBPKT=NBDATA+NBFEC;
    final static int LEFT_DEGREE=3;

    final static int SESSION_TYPE=LDPCFecSession_JNI.TypeSTAIRS;

    final static int SEED=2003;

    final static int VERBOSITY=1;

    final static int DEST_PORT=10978;

    public static void main(String[] args)
    {
	simple_decoder_java();

	return;
    }

    public static int simple_decoder_java()
    {
	// Our session object...
	LDPCFecSession_JNI MyFecSession=new LDPCFecSession_JNI();

	// Received (and rebuilt) packets (DATA and FEC) are stored in a 
	// packets array where each packet is an array of bytes.
	byte[][] packetsArray=null;

	DatagramSocket mySock;
	byte[]	buff	= null;
	int	ret	= -1;
	int	seqNo	= -1;
	int	decodeSteps = 0;
	
	DatagramPacket dp;
	int     i;

	// Sets the verbosity level
	MyFecSession.SetVerbosity(VERBOSITY);

	// Initialize the LDPC session
	if(MyFecSession.InitSession( NBDATA, NBFEC, PKTSZ, LDPCFecSession_JNI.FLAG_DECODER, LEFT_DEGREE, SEED, SESSION_TYPE ) == LDPCFecSession_JNI.LDPC_ERROR)
	{
		System.out.print("Error: Unable to initialize LDPC Session\n");

		MyFecSession=null;

		return -1;
	}
	
	// Initialize our UDP socket

	try
	    {
		mySock = new DatagramSocket(DEST_PORT,InetAddress.getByName("127.0.0.1"));
	    }
	catch (Exception e)
	    {
		System.out.print( "Error initializing socket!\n" );

		MyFecSession.EndSession();
		MyFecSession=null;

		return -1;
	    }

	try
	    {
		buff		= new byte[PKTSZ+4];
		packetsArray	= new byte[NBPKT][];
	    }
	catch (OutOfMemoryError e)
	    {
		System.out.print("Error: insufficient memory (new failed in simple_decoder_java())\n");

		MyFecSession.EndSession();

		MyFecSession=null;
		mySock=null;
		buff=null;
		packetsArray=null;

		return -1;
	    }

	System.out.print( "Decoding in progress...\nWaiting for new packets...\n" );
	while( !MyFecSession.IsDecodingComplete(packetsArray) )
	{
		decodeSteps++;

		try
		    {
			dp=new DatagramPacket(buff,buff.length);

			mySock.receive(dp);
		    }
		catch(IOException e)
		    {
			System.out.print("Error receiving packet!\n" );

			MyFecSession.EndSession();
			
			MyFecSession=null;
			mySock=null;
			buff=null;
			packetsArray=null;
			dp=null;
			
			return -1;
		    }
		catch(OutOfMemoryError e)
		    {
			System.out.print("Error: insufficient memory (new failed for DatagramPacket)\n");

			MyFecSession.EndSession();
			
			MyFecSession=null;
			mySock=null;
			buff=null;
			packetsArray=null;
			dp=null;
			
			return -1;
		    }

		if( dp.getLength() != PKTSZ+4)
		    {
		    
			System.out.print("Error receiving packet!\n");
			
			MyFecSession.EndSession();

			MyFecSession=null;
			mySock=null;
			buff=null;
			packetsArray=null;
			dp=null;
			
			return -1;
		    }

		// OK, new packet received...

		seqNo = (int) buff[0] << 24;
		seqNo+= (int) buff[1] << 16;
		seqNo+= (int) buff[2] << 8;
		seqNo+= (int) buff[3];

		// seqNo = ntohl(*(int*)buff);
		
		System.out.print("------------------------------------\n");
		System.out.print("--- Step "+decodeSteps+" : new packet received: "+seqNo+"\n");

		// Give this new packet to FEC object (will do the needed mem alloc)
		// Fec decoder may recover some missing packets during this call


		// Shift array for DecodeFecStep

		byte[] buff_at_4;

		try
		    {
			buff_at_4=new byte[PKTSZ];
		    }

		catch(OutOfMemoryError e)
		    {
			System.out.print("Error: insufficient memory (new failed for DatagramPacket)\n");
			    
			MyFecSession.EndSession();

			MyFecSession=null;
			mySock=null;
			buff=null;
			packetsArray=null;
			dp=null;
				
			return -1;
		    }

		for(i=0;i<PKTSZ;i++)
		    buff_at_4[i]=buff[i+4];

		MyFecSession.DecodeFecStep( packetsArray, buff_at_4, seqNo, true, 0);
	}

	if (VERBOSITY > 1) // Dump the DATA to screen (if needed)
	{
	    for(i=0; i<NBDATA; i++)
		{
		    System.out.print("DATA["+i+"]= ");
		    DumpBuffer(packetsArray[i]);
		}
	}

	System.out.print("Done! All DATA packets rebuilt in "+decodeSteps+" decoding steps ["+NBDATA+"-"+NBPKT+"]\n");

	mySock.close();

	MyFecSession.EndSession();

	MyFecSession=null;
	mySock=null;
	buff=null;
	packetsArray=null;
	dp=null;

	return ret;
    }

    public static void DumpBuffer(byte[] buff)
    {
	int i;
	
	System.out.print("0x");
	
	for (i=0;i<buff.length;i++)
	    System.out.print(""+buff[i]+":");

	System.out.print("\n");
    }

}

